/*
emonTH temperature and pulse meter code for Coheat monitoring stations
======================================================================

 emonTH temperature node
 ============================================
 
 Ambient humidity & temperature (DHT22 on-board)
 Multiple remote temperature (DS18B20) 
 One pulse meter (connected to terminal block port 4 (IRQ 0 / Digital 2)
 
 emonTx V3 Terminal block:
 1: 5V (from wall wart and to supply water pulse meter)
 2: not used
 3: Ground (from wall wart and to everything)
 4: Input for interrupt from pulse water meter
 5: Usually used to power DS18B20s intermittently; we don't need to do this because we aren't running from battery so can power them from 5V)
 6: Input for DS18B20 data
 
 Provides the following inputs to emonCMS:
 

 1. Pulses since previous transmit
 2. Total Pulses in previous 5 transmits for error check
 3. Flow rate in period (ml/sec)
 4. Consumption in period (ml)
 3. External temperature 1 (first DS18B20)
 4. External temperature 2 (second DS18B20) 
 5. ... and so on. Should automatically detect any DS18B20 connected to the one wire bus. (Up to 60 sensors ordered by address.)
 
 Notes - If you connect additional DS18B20 sensors after node has been set up, the sensor order may change. 
       - Check your inputs in emoncms after adding additional sensors and reassign feeds accordingly.
       -
       - This sketch will also draw more power with more connected sensors and is optimised for getting good data rather than low power consumption.
       - If you have a large number of sensors it is reccomended that you run your node off an external power supply. 
 
 -----------------------------------------------------------------------------------------------------------  
 Technical hardware documentation wiki: http://wiki.openenergymonitor.org/index.php?title=EmonTH
 
 Part of the openenergymonitor.org project
 Licence: GNU GPL V3
 
 Author: James Verrill
 Authors: Dave McCraw (original creator), Marshall Scholz (added autimatic onewire scanning)
 
 Based on the emonTH_DHT22_multiple_DS18B20 sketch by Dave McCraw and Marshall Scholz and the emonTH_PulseCounting sketch by Glyn Hudson and Trystan Lea.
 
 THIS SKETCH REQUIRES:
 
 Libraries in the standard arduino libraries folder:
   - RFu JeeLib           https://github.com/openenergymonitor/RFu_jeelib   - to work with CISECO RFu328 module
   - OneWire library      http://www.pjrc.com/teensy/td_libs_OneWire.html   - DS18B20 sensors
   - DallasTemperature    http://download.milesburton.com/Arduino/MaximTemperature/DallasTemperature_LATEST.zip - DS18B20 sensors
   
 Other files in the project directory 
   - emontx_lib.ino
 */
 
#include <avr/power.h>
#include <avr/sleep.h>
#include <RFu_JeeLib.h>                                                 
#include <OneWire.h>
#include <DallasTemperature.h>
#include <avr/wdt.h> 

 
/*
 Network configuration
 =======================================================================================================================================
 
  - RFM12B frequency can be RF12_433MHZ, RF12_868MHZ or RF12_915MHZ. You should use the one matching the module you have.
  - RFM12B wireless network group - needs to be same as emonBase and emonGLCD
  - RFM12B node ID - should be unique on network
 
 Recommended node ID allocation
 ------------------------------------------------------------------------------------------------------------
 -ID-	-Node Type- 
 0	- Special allocation in JeeLib RFM12 driver - reserved for OOK use
 1-4    - Control nodes 
 5-10	- Energy monitoring nodes
 11-14	--Un-assigned --
 15-16	- Base Station & logging nodes
 17-30	- Environmental sensing nodes (temperature humidity etc.)
 31	- Special allocation in JeeLib RFM12 driver - Node31 can communicate with nodes on any network group
 -------------------------------------------------------------------------------------------------------------
 */
#define FREQUENCY RF12_433MHZ   // Transmitter frequency. Only one should be uncommented at a time.
//#define FREQUENCY RF12_915MHZ 
//#define FREQUENCY RF12_868MHZ 
const int NETWORK_GROUP = 210;  // Default 210
const int NODE_ID       = 19;   // Dafault 19
const int UNO = 1;                                                      // Set to 0 if your not using the UNO bootloader (i.e using Duemilanove) - All Atmega's shipped from OpenEnergyMonitor come with Arduino Uno bootloader


// Monitoring configuration
// ======================================================================================================================================
 const int mSECS_BETWEEN_READINGS = 10000;  // Default = "10000" = 10 seconds. How long to wait between readings, in milliseconds.
 

// emonTH pin allocations 
// ======================================================================================================================================
const int LED          = 9;  // Default 9


// Onewire bus configaturation
// ======================================================================================================================================
const int DS18B20_PWR  = 5;  // Default 5
const int ONE_WIRE_BUS = 19;  // Default 19

const int ASYNC_DELAY           = 375; // Default 375 - Delay for onewire sensors to respond
const int TEMPERATURE_PRECISION = 12;  // Default 12  - Onewire temperature sensor precisionn. Details found below.
 /*
  NOTE: - There is a trade off between power consumption and sensor resolution.
        - A higher resolution will keep the processor awake longer - Approximate values found below.
                                     
  - DS18B20 temperature precision:
      9bit: 0.5C,  10bit: 0.25C,  11bit: 0.1125C, 12bit: 0.0625C
  - Required delay when reading DS18B20
      9bit: 95ms,  10bit: 187ms,  11bit: 375ms,   12bit: 750ms
     
  More info can be found here: http://harizanov.com/2013/07/optimizing-ds18b20-code-for-low-power-applications/   
 */


// Pulse Counter Configuration
// ======================================================================================================================================

const int pulseIRQport = 0; // Default is 0 to match emonTX IRQ 0 = Digital 2 hardwired to emonTx terminal block port 4.
const int pulseTotalNumber = 5; // number of pulse readings to record and send in the total number for error checking. default is 5.

// variables for freq to flow rate conversion
const float squareConst = 0.0107; 
const float linearConst = 6.7506;

// end of configuration
// ======================================================================================================================================
// ======================================================================================================================================
// Start of part that does stuff




// Global variables
boolean debug; // variable to store is debug is avalable or not. 
int numberOfDevices; // Number of temperature devices found
DeviceAddress tempDeviceAddress; // We'll use this variable to store a found onewire device address


//uint8_t MARKO COMMENTED THIS OUT BECAUSE CODE WOULDN'T WORK OTHERWISE
int PayloadLength = 8; // initial, non variable length in bytes. 2 bytes per int / variable in array
                       // The default "8" currently covers the pulse variables and flow variables
                       // This will be incremented according to the number of onewire sensors on the bus.

 
int pulseCount = 0;                                                    // Number of pulses, used to measure energy.
int lastPulseCount = 0;
int pulseReadings[pulseTotalNumber] = {0, 0, 0, 0, 0};
unsigned long systemTimeLastCount = 0;   // in milliseconds

 #define MaxOnewire 60  // Maximum number of sensors on the onewire bus - too big of a number may create too big of a data packet (max packet size is 128 bytes) - defiult "60" to allow for (almost) full 128 byte packet length. 61 would make it full
 
// RFM12B RF payload datastructure
typedef struct {  // must be kept to less than 128 bytes     
  int lastPulses;  // 2 bytes for each int
  int lastPulsesTotal;  // 4 bytes of 126 byte(maximum) rf packet used for all these. this number goes into the "PayloadLength" inital value above   	                                      
  int flowRate;
  int consumption;
  int onewireTemp[MaxOnewire];	  // 2 additional bytes used for each sensor - maximum of 61 sensors supported with this arrangement and rf packet type. (If you had to hook up more than 60 sensors to one node, I pity you.)
} Payload_t; // create datatype payload

  Payload_t rfPayload; // make a new variable "rfPayload" with type "payload"'



// Attach JeeLib sleep function to Atmega328 watchdog - enables MCU to be put into sleep mode between readings to reduce power consumption 
ISR(WDT_vect) { 
  //Sleepy::watchdogEvent(); 
} 


// OneWire for DS18B20
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);

 
 

/**
 * setup() - called once on boot to initialise the emonTH
 */
void setup() {

  // Output only if serial UART to USB is connected
  debug = Serial ? 1 : 0;                              

  print_welcome_message();  
  set_pin_modes();
  
  // LED on
  digitalWrite(LED, HIGH);

  // Initialize RFM12B
  rf12_initialize(NODE_ID, FREQUENCY, NETWORK_GROUP);                       
  //rf12_sleep(RF12_SLEEP);

  // turn off non crital chip functions to reduce battery drain
  //reduce_power();
  
  // Find & Initialise sensors
  initialise_DS18B20();
  initialise_PulseCount();
  // Confirm we've got at least one sensor, or shut down. this will only check for temp sensors(!)
  validate_sensor_presence();

  // LED off
  digitalWrite(LED, LOW);

} // end of setup




/**
 * Perform temperature and humidity sending
 */
void loop()
{ 

  if(debug)
    Serial.println("PROGRESS - temp radings");
    
  // External temperature readings
  if (numberOfDevices >= 0) 
    take_ds18b20_reading();

  if(debug)
    Serial.println("PROGRESS - pulse reading");
  
  // Pulse counter reading
  //take_pulseReading();
  //take_pulseCount();
  take_flowReading();
  
  if(debug)
    Serial.println("PROGRESS - prep radio");
  
  // Debugging
  print_payload();                                             

  if(debug)
    Serial.println("PROGRESS - radio transmit");

  // power up radio and send data packet
  power_spi_enable();  
  //rf12_sleep(RF12_WAKEUP);
  rf12_sendNow(0, &rfPayload, PayloadLength); // new sender that only sends used variables calculated in onewire initilisation. 
  rf12_sendWait(2);
  //rf12_sleep(RF12_SLEEP);
  power_spi_disable();  

  if(debug)
    Serial.println("PROGRESS - flash led");


  if (debug){
    flash_led(50);
  }

  if(debug)
    Serial.println("PROGRESS - wait 10 secs");


  // That's it - wait until next time :)
  dodelay(mSECS_BETWEEN_READINGS);
  if(debug)
    Serial.println("PROGRESS - looping");
  
  
}

/**
 * Monitor for pulses
 */
void onPulse()                  
{
  pulseCount++;  
}




//////////////////////////////////////////////////
/**
 * Set the pin modes required for this sketch
 */
void set_pin_modes()
{
  pinMode(LED,         OUTPUT); 
  pinMode(DS18B20_PWR, OUTPUT);
}




/////////////////////////////////////////////////
/**
 * Flash the LED for the stated period
 */
void flash_led (int duration){
  digitalWrite(LED,HIGH);
  dodelay(duration);
  digitalWrite(LED,LOW); 
}


//////////////////////////////////////////////////
/**
 * Find the expected DS18B20 sensors
 *
 * Automatically scans the entire onewire bus for sensors and stores their adressess in the device adresses array.
 * 
 * Should support up to 60 sensors. (limited by size of data packet) Currently tested up to four sensors.
 */
void initialise_DS18B20()
{
  // Switch on - ignore as we constantly power
  digitalWrite(DS18B20_PWR, HIGH); 
  dodelay(50); 

  // start up oonewire library
  sensors.begin();

  // Disable automatic temperature conversion to reduce time spent awake, instead we sleep for ASYNC_DELAY
  // see http://harizanov.com/2013/07/optimizing-ds18b20-code-for-low-power-applications/ 
  sensors.setWaitForConversion(false);                             

  // Grab a count of devices on the wire
  numberOfDevices = sensors.getDeviceCount();
 
 // Detect if too many devices are connected 
  if (numberOfDevices > MaxOnewire) 
  { 
    numberOfDevices = MaxOnewire; // Set number of devices to maximum allowed sensors to prevent the extra sensors from being included.
  }
  
 // add bytes that the sensors will use onto the payloadlength. Every sensor is a 16  bit integer that will take up 2 bytes
  PayloadLength = PayloadLength + (numberOfDevices * 2);
  
  if (PayloadLength >= 128) // check if payload length is out of acceptable values
  {
    PayloadLength = 128; // set length to maximim acceptable value
  }
 
 // bus info
  if (debug){
    
     // locate devices on the bus
  Serial.print("Locating devices...");
  Serial.print("Found ");
  Serial.print(numberOfDevices, DEC);
  Serial.println(" devices.");
  
  if (numberOfDevices > MaxOnewire) {
    Serial.println("Too many devices!!!");
    Serial.print("maximum allowed number of devices is ");
    Serial.println(MaxOnewire);
  }

  // report parasite power requirements
  Serial.print("Parasite power is: "); 
  if (sensors.isParasitePowerMode()) Serial.println("ON");
  else Serial.println("OFF");
  } 
 
 
 // Loop through each device, print out address
  for(int i=0;i<numberOfDevices; i++)
  {
    // Search the wire for address
    if(sensors.getAddress(tempDeviceAddress, i))
	{
		if (debug) {
                  Serial.print("Found device ");
		  Serial.print(i, DEC);
		  Serial.print(" with address: ");
		  printAddress(tempDeviceAddress); // just prints sensor adress on serial bus
		  Serial.println();
		
		  Serial.print("Setting resolution to ");
		  Serial.println(TEMPERATURE_PRECISION, DEC);
                }
		
		// set the resolution to TEMPERATURE_PRECISION bit (Each Dallas/Maxim device is capable of several different resolutions)
		sensors.setResolution(tempDeviceAddress, TEMPERATURE_PRECISION);
		
		if (debug) {
		  Serial.print("Resolution actually set to: ");
		  Serial.print(sensors.getResolution(tempDeviceAddress), DEC); 
		  Serial.println();
                }
	}else{
              
	      if (debug) {
		Serial.print("Found ghost device at ");
                Serial.print(i, DEC);
		Serial.println(" but could not detect address. Check power and cabling");
              }
	}
  }

  // Switch off for now - not used. powered permenantly
  //digitalWrite(DS18B20_PWR, LOW);
  
  Serial.print(" rf packet length calculated at: ");
  Serial.println( PayloadLength );
}

// function to print a device address
void printAddress(DeviceAddress deviceAddress)
{
  for (uint8_t i = 0; i < 8; i++)
  {
    if (deviceAddress[i] < 16) Serial.print("0");
    Serial.print(deviceAddress[i], HEX);
  }
}




//////////////////////////////////////////////////
/** 
 * If we don't have at least one sensor available, it's sleep time
 */
void validate_sensor_presence()
{
  if (!numberOfDevices >= 1) 
  {
    if (debug) {
      
      Serial.print("Power down - no sensors detected at all!");
    }
    
    for (int i=0; i<10; i++)
    {
      flash_led(250); 
      dodelay(250);
    }
    //cli();                                      //stop responding to interrupts - ignore, we still use flow sensor
    //Sleepy::powerDown();                        //sleep forever - ignore, we still use flow sensor
  }
}


//////////////////////////////////////////////////
/**
 * Convenience method; read from all DS18B20s
 * 
 */
void take_ds18b20_reading () 
{
  // Power up - not used
  // digitalWrite(DS18B20_PWR, HIGH); 
  // dodelay(50); 
  
  
  // call sensors.requestTemperatures() to issue a global temperature 
  // request to all devices on the bus
  sensors.requestTemperatures(); // Send the command to get temperatures
  Serial.println("DONE");
  
  
  // Loop through each device, print out temperature data
  for(int i=0;i<numberOfDevices; i++)
  {
    // Search the wire for address
    if(sensors.getAddress(tempDeviceAddress, i))
      {	
           sensors.setResolution(tempDeviceAddress, TEMPERATURE_PRECISION);
           
           // Get readings. We must wait for ASYNC_DELAY due to power-saving (waitForConversion = false)
           sensors.requestTemperatures();                                   
           dodelay(ASYNC_DELAY); 
           float temp1=(sensors.getTempC(tempDeviceAddress));
           
           // Payload will maintain previous reading unless the temperature is within range.
           if (temperature_in_range(temp1)){
           rfPayload.onewireTemp[i] = temp1 * 10; // add sensor temp to rfpayload in onewireTemp array
	   } 
	//else ghost device! Check your power requirements and cabling
      }
  }

  // Power down - not used
  // digitalWrite(DS18B20_PWR, LOW);
 
}


//////////////////////////////////////////////////
/**
 * validate that the provided temperature is within acceptable bounds.
 */
boolean temperature_in_range(float temp)
{
  // Only accept the reading if it's within a desired range.
  float minimumTemp = -40.0;
  float maximumTemp = 125.0;

  return temp > minimumTemp && temp < maximumTemp;
}


//////////////////////////////////////////////////
/**
 * Sets up the pulse counting module 
 */
 void initialise_PulseCount()
 {
  attachInterrupt(pulseIRQport, onPulse, FALLING);  // set up pulse count
  //if (UNO) wdt_enable(WDTO_8S);
  if (!debug)
    return;
  
  Serial.println("Pulse count initialised");
 }
 
 //////////////////////////////////////////////////
/**
 * Convenience method, record a pulse
 * 
 */
void countPulse()                  
{
  pulseCount++;              
  if(debug) {
    Serial.print("p");
  }
}

void take_pulseCount() {
  rfPayload.lastPulses = pulseCount;
  
}


 //////////////////////////////////////////////////
/**
 * Records number of pulses into 
 * 
 */
void take_pulseReading() {
  // define local variables
  int topRange;
  int bottomRange;
  int thisPulseCount;
  
  // record absolute pulse count at time of reading start of operation // safety from interupts mid flow
  thisPulseCount = pulseCount;
  
  // forget the most historic pulse reading
  for(int i = 0; i < (pulseTotalNumber - 1); i++) {
    pulseReadings[i] = pulseReadings[i+1];
  }
  // calculate this pulse reading and reset lastPulseCount
  if (thisPulseCount >= lastPulseCount) {
      pulseReadings[(pulseTotalNumber - 1)] = thisPulseCount - lastPulseCount;
  }
  else {
      topRange = 32767 - lastPulseCount; // calculate amount before unsigned long variable rollover
      bottomRange = thisPulseCount + 32768;
      pulseReadings[(pulseTotalNumber - 1)] = bottomRange + topRange;             
  }
  
  // calculate sum of last pulses
  int sumPulses = 0;
  for(int i = 0; i <= (pulseTotalNumber - 1); i++) {
    sumPulses += pulseReadings[i];
  }
   
  // reset last pulse count
  lastPulseCount = thisPulseCount;
    
  // add compontents to rf message
  rfPayload.lastPulses = pulseReadings[(pulseTotalNumber - 1)];
  rfPayload.lastPulsesTotal = sumPulses;
  
  if(!debug)
    return;
    
  Serial.println("PULSE COUNT");
  Serial.print("Pulse counter: ");
  Serial.println(thisPulseCount); 
  Serial.print("Latest pulse reading: ");
  Serial.println(pulseReadings[(pulseTotalNumber - 1)]);
  Serial.print("Latest pulse sum: ");
  Serial.print(pulseTotalNumber);
  Serial.print(" readings ");
  Serial.println(sumPulses);
 
}


 //////////////////////////////////////////////////
/**
 * Calculates flow rate on the pulse data stored in rfpayload. Also calls for these to be updated.
 * 
 */
void take_flowReading() {
  // define local variables
  unsigned long systemTimeThisCount, systemTimeDifference;
  
  int pulses;
  float freq, flowRate;  // litres per second
  float millilitres;
  
  // record the absolute time difference since the last message send
  systemTimeThisCount = millis();
  systemTimeDifference = systemTimeThisCount - systemTimeLastCount;  // in milliseconds  

  take_pulseReading();
  pulses = pulseReadings[(pulseTotalNumber - 1)];
  
  // calculate average flow rate for time period
  freq = (float)pulses / (float)systemTimeDifference * 1000;    // convert to Hz
  
  // covert to flow rate   ml/min
  flowRate = (sq(freq)*squareConst + freq*linearConst) *1000 / 60; // (flowrate L/H * 1000 ml/L / 60 min/hr)
        
  // translate flow rate into actual millilitres in time period
  millilitres = flowRate / 60 * systemTimeDifference / 1000 ;  // (flowrate ml/min / 60s/min * time (ms) / 1000 ms/s )
  
  // reset system time memory
  systemTimeLastCount = systemTimeThisCount;
  
  // add compontents to rf message
  rfPayload.flowRate = (int)(flowRate);
  rfPayload.consumption = (int)millilitres;
  
  if(!debug)
    return;
    
  Serial.println("FLOW CALCULATION");
  Serial.print("Frequency: ");
  Serial.println(freq); 
  Serial.print("FlowRate: ");
  Serial.println(flowRate); 
  Serial.print("Consumption: ");
  Serial.println(millilitres); 
  Serial.print("System Time: ");
  Serial.println(systemTimeThisCount); 
  Serial.print("System Time Difference: ");
  Serial.println(systemTimeDifference); 

}

//////////////////////////////////////////////////
/**
 * For debugging purposes: print the payload as it will shortly be sent to the emonBASE
 * Wise to extend this if you have extra sensors wired in.
 */
void print_payload()
{
  if (!debug)
    return;
  
  Serial.println("emonTH payload: ");
  Serial.print(" Last pulses: ");
  Serial.print(rfPayload.lastPulses);
  Serial.print(" Last " );
  Serial.print(pulseTotalNumber);
  Serial.print(" pulses total: ");
  Serial.print(rfPayload.lastPulsesTotal);
  

  
  if (numberOfDevices >= 1){
    for(int i=0;i<numberOfDevices; i++)
      {
        Serial.print("sensor ");
        Serial.print(i);
        Serial.print(" = ");
        Serial.print( rfPayload.onewireTemp[i] /10.0); 
        Serial.println("C");
       }
  } else {
    Serial.println("No onewire sensors");
  }
  
  //Serial.print("rfpayload is ");
  //Serial.print(rfPayload); // print the whole rfpayload for software debugging.
  Serial.println();
}

//////////////////////////////////////////////////
/**
 * Dumps useful intro to serial
 */
void print_welcome_message()
{
  if (!debug)
    return;

  delay(50);
  Serial.begin(9600);
  dodelay(10000);
  
  Serial.println("emonTH : OpenEnergyMonitor.org");
  
  Serial.print("Node: "); 
  Serial.print(NODE_ID); 
 
  Serial.print(" Freq: "); 
  switch(FREQUENCY){
  case RF12_433MHZ:
    Serial.print("433Mhz");
    break;
  case RF12_868MHZ:
    Serial.print("868Mhz");
    break;
  case RF12_915MHZ:
    Serial.print("915Mhz");
    break;
    
  }

  
  Serial.print(" Network: "); 
  Serial.println(NETWORK_GROUP);

  
  dodelay(100);
}

//////////////////////////////////////////////////
/**
 * Turn off what we don't need.
 * see http://www.nongnu.org/avr-libc/user-manual/group__avr__power.html
 */
void reduce_power()
{
  //ACSR |= (1 << ACD);              // Disable Analog comparator    
  //power_twi_disable();             // Disable the Two Wire Interface module.

  //power_timer1_disable();          // Timer 1
  //power_spi_disable();             // Serial peripheral interface

  //if (!debug){
  // power_usart0_disable();        // Disable serial UART if not connected
  //}  

  //power_timer0_enable();           // Necessary for the DS18B20 library.
}

//////////////////////////////////////////////////
/**
 * Power-friendly delay
 */
void dodelay(unsigned int ms)
{
  delay(ms);
  // /*
  //byte oldADCSRA=ADCSRA;
  //byte oldADCSRB=ADCSRB;
  //byte oldADMUX=ADMUX;

  //Sleepy::loseSomeTime(ms); // JeeLabs power save function: enter low power mode for x seconds (valid range 16-65000 ms)

  //ADCSRA=oldADCSRA;         // restore ADC state
  //ADCSRB=oldADCSRB;
  //ADMUX=oldADMUX;
  // */
}


